/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
﻿namespace ZooKeeperNet
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.IO;
    using System.Runtime.CompilerServices;
    using log4net;
    using Org.Apache.Zookeeper.Data;
    using Org.Apache.Zookeeper.Proto;

    [DebuggerDisplay("Id = {Id}")]
    public class ZooKeeper : IDisposable, IZooKeeper
    {
        private static readonly ILog LOG = LogManager.GetLogger(typeof(ZooKeeper));

        private readonly ZKWatchManager watchManager = new ZKWatchManager();

        public List<string> DataWatches
        {
            get
            {
                return new List<string>(watchManager.dataWatches.Keys);
            }
        }

        public List<string> ExistWatches
        {
            get
            {
                return new List<string>(watchManager.existWatches.Keys);
            }
        }

        public List<string> ChildWatches
        {
            get
            {
                return new List<string>(watchManager.childWatches.Keys);
            }
        }

        /// <summary>
        /// Manage watchers & handle events generated by the ClientCnxn object.
        ///
        /// We are implementing this as a nested class of ZooKeeper so that
        /// the public methods will not be exposed as part of the ZooKeeper client
        /// API.
        /// </summary>
        public abstract class WatchRegistration
        {
            private readonly IWatcher watcher;
            private readonly string clientPath;

            /// <summary>
            /// Initializes a new instance of the <see cref="WatchRegistration"/> class.
            /// </summary>
            /// <param name="watcher">The watcher.</param>
            /// <param name="clientPath">The client path.</param>
            protected WatchRegistration(IWatcher watcher, string clientPath)
            {
                this.watcher = watcher;
                this.clientPath = clientPath;
            }

            abstract protected Dictionary<string, HashSet<IWatcher>> GetWatches(int rc);

            /// <summary>
            /// Register the watcher with the set of watches on path.
            /// </summary>
            /// <param name="rc">the result code of the operation that attempted to add the watch on the path.</param>
            public void Register(int rc)
            {
                if (!ShouldAddWatch(rc)) return;

                Dictionary<string, HashSet<IWatcher>> watches = GetWatches(rc);
                lock (watches)
                {
                    HashSet<IWatcher> watchers;
                    watches.TryGetValue(clientPath, out watchers);
                    if (watchers == null)
                    {
                        watchers = new HashSet<IWatcher>();
                        watches[clientPath] = watchers;
                    }
                    watchers.Add(watcher);
                }
            }

            /// <summary>
            /// Determine whether the watch should be added based on return code.
            /// </summary>
            /// <param name="rc">the result code of the operation that attempted to add the watch on the node</param>
            /// <returns>true if the watch should be added, otw false</returns>
            protected virtual bool ShouldAddWatch(int rc)
            {
                return rc == 0;
            }
        }

        /// <summary>
        /// Handle the special case of exists watches - they add a watcher
        /// even in the case where NONODE result code is returned.
        /// </summary>
        internal class ExistsWatchRegistration : WatchRegistration
        {
            private readonly ZKWatchManager watchManager;

            public ExistsWatchRegistration(ZKWatchManager watchManager, IWatcher watcher, string clientPath)
                : base(watcher, clientPath)
            {
                this.watchManager = watchManager;
            }

            protected override Dictionary<string, HashSet<IWatcher>> GetWatches(int rc)
            {
                return rc == 0 ? watchManager.dataWatches : watchManager.existWatches;
            }

            protected override bool ShouldAddWatch(int rc)
            {
                return rc == 0 || rc == (int)KeeperException.Code.NONODE;
            }
        }

        internal class DataWatchRegistration : WatchRegistration
        {
            private readonly ZKWatchManager watchManager;

            public DataWatchRegistration(ZKWatchManager watchManager, IWatcher watcher, string clientPath)
                : base(watcher, clientPath)
            {
                this.watchManager = watchManager;
            }

            protected override Dictionary<string, HashSet<IWatcher>> GetWatches(int rc)
            {
                return watchManager.dataWatches;
            }
        }

        internal class ChildWatchRegistration : WatchRegistration
        {
            private readonly ZKWatchManager watchManager;

            public ChildWatchRegistration(ZKWatchManager watchManager, IWatcher watcher, string clientPath)
                : base(watcher, clientPath)
            {
                this.watchManager = watchManager;
            }

            protected override Dictionary<string, HashSet<IWatcher>> GetWatches(int rc)
            {
                return watchManager.childWatches;
            }
        }

        public class States : IEquatable<States>
        {
            public static readonly States CONNECTING = new States("CONNECTING");
            public static readonly States ASSOCIATING = new States("ASSOCIATING");
            public static readonly States CONNECTED = new States("CONNECTED");
            public static readonly States CLOSED = new States("CLOSED");
            public static readonly States AUTH_FAILED = new States("AUTH_FAILED");

            public string state;

            public States(string state)
            {
                this.state = state;
            }

            public string State
            {
                get { return state; }
            }

            public bool IsAlive()
            {
                return this != CLOSED && this != AUTH_FAILED;
            }

            public bool Equals(States other)
            {
                if (ReferenceEquals(null, other)) return false;
                if (ReferenceEquals(this, other)) return true;
                return Equals(other.state, state);
            }

            public override bool Equals(object obj)
            {
                if (ReferenceEquals(null, obj)) return false;
                if (ReferenceEquals(this, obj)) return true;
                if (obj.GetType() != typeof(States)) return false;
                return Equals((States)obj);
            }

            public override int GetHashCode()
            {
                return (state != null ? state.GetHashCode() : 0);
            }
        }

        protected Guid id = Guid.NewGuid();
        protected volatile States state;
        protected IClientConnection cnxn;

        /// <summary>
        /// To create a ZooKeeper client object, the application needs to pass a
        /// connection string containing a comma separated list of host:port pairs,
        /// each corresponding to a ZooKeeper server.
        /// 
        /// Session establishment is asynchronous. This constructor will initiate
        /// connection to the server and return immediately - potentially (usually)
        /// before the session is fully established. The watcher argument specifies
        /// the watcher that will be notified of any changes in state. This
        /// notification can come at any point before or after the constructor call
        /// has returned.
        /// 
        /// The instantiated ZooKeeper client object will pick an arbitrary server
        /// from the connectstring and attempt to connect to it. If establishment of
        /// the connection fails, another server in the connect string will be tried
        /// (the order is non-deterministic, as we random shuffle the list), until a
        /// connection is established. The client will continue attempts until the
        /// session is explicitly closed.
        /// 
        /// Added in 3.2.0: An optional "chroot" suffix may also be appended to the
        /// connection string. This will run the client commands while interpreting
        /// all paths relative to this root (similar to the unix chroot command).
        /// </summary>
        /// <param name="connectstring">
        ///            comma separated host:port pairs, each corresponding to a zk
        ///            server. e.g. "127.0.0.1:3000,127.0.0.1:3001,127.0.0.1:3002" If
        ///            the optional chroot suffix is used the example would look
        ///            like: "127.0.0.1:3000,127.0.0.1:3001,127.0.0.1:3002/app/a"
        ///            where the client would be rooted at "/app/a" and all paths
        ///            would be relative to this root - ie getting/setting/etc...
        ///            "/foo/bar" would result in operations being run on
        ///            "/app/a/foo/bar" (from the server perspective).
        /// </param>
        /// <param name="sessionTimeout">The session timeout.</param>
        /// <param name="watcher">
        ///            a watcher object which will be notified of state changes, may
        ///            also be notified for node events
        /// </param>
        public ZooKeeper(string connectstring, TimeSpan sessionTimeout, IWatcher watcher)
        {
            LOG.Info(string.Format("Initiating client connection, connectstring={0} sessionTimeout={1} watcher={2}", connectstring, sessionTimeout, watcher));

            watchManager.defaultWatcher = watcher;
            cnxn = new ClientConnection(connectstring, sessionTimeout, this, watchManager);
            cnxn.Start();
        }

        public ZooKeeper(string connectstring, TimeSpan sessionTimeout, IWatcher watcher, long sessionId, byte[] sessionPasswd)
        {
            LOG.Info(string.Format("Initiating client connection, connectstring={0} sessionTimeout={1} watcher={2} sessionId={3} sessionPasswd={4}", connectstring, sessionTimeout, watcher, sessionId, (sessionPasswd == null ? "<null>" : "<hidden>")));

            watchManager.defaultWatcher = watcher;
            cnxn = new ClientConnection(connectstring, sessionTimeout, this, watchManager, sessionId, sessionPasswd);
            cnxn.Start();
        }

        /// <summary>
        /// Unique ID representing the instance of the client
        /// </summary>
        public Guid Id
        {
            get { return id; }
            set { id = value; }
        }

        /// <summary>
        /// The session id for this ZooKeeper client instance. The value returned is
        /// not valid until the client connects to a server and may change after a
        /// re-connect.
        /// </summary>
        /// <value>The session id.</value>
        public long SessionId
        {
            get
            {
                return cnxn.SessionId;
            }
        }

        /// <summary>
        /// The session password for this ZooKeeper client instance. The value
        /// returned is not valid until the client connects to a server and may
        ///  change after a re-connect.
        ///
        /// This method is NOT thread safe
        /// </summary>
        /// <value>The sesion password.</value>
        public byte[] SesionPassword
        {
            get
            {
                return cnxn.SessionPassword;
            }
        }

        /// <summary>
        /// The negotiated session timeout for this ZooKeeper client instance. The
        /// value returned is not valid until the client connects to a server and
        /// may change after a re-connect.
        /// 
        /// This method is NOT thread safe
        /// </summary>
        /// <value>The session timeout.</value>
        public TimeSpan SessionTimeout
        {
            get
            {
                return cnxn.SessionTimeout;
            }
        }

        /// <summary>
        /// Add the specified scheme:auth information to this connection.
        ///
        /// This method is NOT thread safe
        /// </summary>
        /// <param name="scheme">The scheme.</param>
        /// <param name="auth">The auth.</param>
        public void AddAuthInfo(string scheme, byte[] auth)
        {
            cnxn.AddAuthInfo(scheme, auth);
        }

        /// <summary>
        /// Specify the default watcher for the connection (overrides the one
        /// specified during construction).
        /// </summary>
        /// <param name="watcher">The watcher.</param>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void Register(IWatcher watcher)
        {
            watchManager.defaultWatcher = watcher;
        }

        public States State
        {
            get
            {
                return state;
            }
            internal set
            {
                state = value;
            }
        }

        /// <summary>
        /// Close this client object. Once the client is closed, its session becomes
        /// invalid. All the ephemeral nodes in the ZooKeeper server associated with
        /// the session will be removed. The watches left on those nodes (and on
        /// their parents) will be triggered.
        /// </summary>   
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void Dispose()
        {
            if (!state.IsAlive())
            {
                if (LOG.IsDebugEnabled)
                {
                    LOG.Debug("Close called on already closed client");
                }
                return;
            }

            if (LOG.IsDebugEnabled)
            {
                LOG.Debug(string.Format("Closing session: 0x{0:X}", SessionId));
            }

            try
            {
                cnxn.Dispose();
            }
            catch (IOException e)
            {
                if (LOG.IsDebugEnabled)
                {
                    LOG.Debug("Ignoring unexpected exception during close", e);
                }
            }

            LOG.Info(string.Format("Session: 0x{0:X} closed", SessionId));
        }

        /// <summary>
        /// Prepend the chroot to the client path (if present). The expectation of
        /// this function is that the client path has been validated before this
        /// function is called
        /// </summary>
        /// <param name="clientPath">The path to the node.</param>
        /// <returns>server view of the path (chroot prepended to client path)</returns>
        private string PrependChroot(string clientPath)
        {
            if (cnxn.ChrootPath != null)
            {
                // handle clientPath = "/"
                return clientPath.Length == 1 ? cnxn.ChrootPath : cnxn.ChrootPath + clientPath;
            }
            return clientPath;
        }

        /// <summary>
        /// Create a node with the given path. The node data will be the given data,
        /// and node acl will be the given acl.
        /// 
        /// The flags argument specifies whether the created node will be ephemeral
        /// or not.
        /// 
        /// An ephemeral node will be removed by the ZooKeeper automatically when the
        /// session associated with the creation of the node expires.
        /// 
        /// The flags argument can also specify to create a sequential node. The
        /// actual path name of a sequential node will be the given path plus a
        /// suffix "i" where i is the current sequential number of the node. The sequence
        /// number is always fixed length of 10 digits, 0 padded. Once
        /// such a node is created, the sequential number will be incremented by one.
        /// 
        /// If a node with the same actual path already exists in the ZooKeeper, a
        /// KeeperException with error code KeeperException.NodeExists will be
        /// thrown. Note that since a different actual path is used for each
        /// invocation of creating sequential node with the same path argument, the
        /// call will never throw "file exists" KeeperException.
        /// 
        /// If the parent node does not exist in the ZooKeeper, a KeeperException
        /// with error code KeeperException.NoNode will be thrown.
        /// 
        /// An ephemeral node cannot have children. If the parent node of the given
        /// path is ephemeral, a KeeperException with error code
        /// KeeperException.NoChildrenForEphemerals will be thrown.
        /// 
        /// This operation, if successful, will trigger all the watches left on the
        /// node of the given path by exists and getData API calls, and the watches
        /// left on the parent node by getChildren API calls.
        /// 
        /// If a node is created successfully, the ZooKeeper server will trigger the
        /// watches on the path left by exists calls, and the watches on the parent
        /// of the node by getChildren calls.
        /// 
        /// The maximum allowable size of the data array is 1 MB (1,048,576 bytes).
        /// Arrays larger than this will cause a KeeperExecption to be thrown.
        /// </summary>
        /// <param name="path">The path for the node.</param>
        /// <param name="data">The data for the node.</param>
        /// <param name="acl">The acl for the node.</param>
        /// <param name="createMode">specifying whether the node to be created is ephemeral and/or sequential.</param>
        /// <returns></returns>
        public string Create(string path, byte[] data, List<ACL> acl, CreateMode createMode)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath, createMode.Sequential);

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.Create;
            CreateRequest request = new CreateRequest();
            CreateResponse response = new CreateResponse();
            request.Data = data;
            request.Flags = createMode.Flag;
            request.Path = serverPath;
            if (acl != null && acl.Count == 0)
            {
                throw new KeeperException.InvalidACLException();
            }
            request.Acl = acl;
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, null);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            return cnxn.ChrootPath == null ? response.Path : response.Path.Substring(cnxn.ChrootPath.Length);
        }

        /// <summary>
        /// Delete the node with the given path. The call will succeed if such a node
        /// exists, and the given version matches the node's version (if the given
        /// version is -1, it matches any node's versions).
        ///
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if the nodes does not exist.
        ///
        /// A KeeperException with error code KeeperException.BadVersion will be
        /// thrown if the given version does not match the node's version.
        ///
        /// A KeeperException with error code KeeperException.NotEmpty will be thrown
        /// if the node has children.
        /// 
        /// This operation, if successful, will trigger all the watches on the node
        /// of the given path left by exists API calls, and the watches on the parent
        /// node left by getChildren API calls.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="version">The version.</param>
        public void Delete(string path, int version)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            string serverPath;

            // maintain semantics even in chroot case
            // specifically - root cannot be deleted
            // I think this makes sense even in chroot case.
            if (clientPath.Equals("/"))
            {
                // a bit of a hack, but delete(/) will never succeed and ensures
                // that the same semantics are maintained
                serverPath = clientPath;
            }
            else
            {
                serverPath = PrependChroot(clientPath);
            }

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.Delete;
            DeleteRequest request = new DeleteRequest();
            request.Path = serverPath;
            request.Version = version;
            ReplyHeader r = cnxn.SubmitRequest(h, request, null, null);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
        }

        /// <summary>
        /// Return the stat of the node of the given path. Return null if no such a
        /// node exists.
        /// 
        /// If the watch is non-null and the call is successful (no exception is thrown),
        /// a watch will be left on the node with the given path. The watch will be
        /// triggered by a successful operation that creates/delete the node or sets
        /// the data on the node.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="watcher">The watcher.</param>
        /// <returns>the stat of the node of the given path; return null if no such a node exists.</returns>
        public Stat Exists(string path, IWatcher watcher)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            // the watch contains the un-chroot path
            WatchRegistration wcb = null;
            if (watcher != null)
            {
                wcb = new ExistsWatchRegistration(watchManager, watcher, clientPath);
            }

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.Exists;
            ExistsRequest request = new ExistsRequest();
            request.Path = serverPath;
            request.Watch = watcher != null;
            SetDataResponse response = new SetDataResponse();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, wcb);
            if (r.Err != 0)
            {
                if (r.Err == (int)KeeperException.Code.NONODE)
                {
                    return null;
                }
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }

            return response.Stat.Czxid == -1 ? null : response.Stat;
        }

        /// <summary>
        /// Return the stat of the node of the given path. Return null if no such a
        /// node exists.
        /// 
        /// If the watch is true and the call is successful (no exception is thrown),
        /// a watch will be left on the node with the given path. The watch will be
        /// triggered by a successful operation that creates/delete the node or sets
        /// the data on the node.
        /// @param path
        ///                the node path
        /// @param watch
        ///                whether need to watch this node
        /// @return the stat of the node of the given path; return null if no such a
        ///         node exists.
        /// @throws KeeperException If the server signals an error
        /// @throws InterruptedException If the server transaction is interrupted.
        /// </summary>
        public Stat Exists(string path, bool watch)
        {
            return Exists(path, watch ? watchManager.defaultWatcher : null);
        }

        /// <summary>
        /// Return the data and the stat of the node of the given path.
        /// 
        /// If the watch is non-null and the call is successful (no exception is
        /// thrown), a watch will be left on the node with the given path. The watch
        /// will be triggered by a successful operation that sets data on the node, or
        /// deletes the node.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// @param path the given path
        /// @param watcher explicit watcher
        /// @param stat the stat of the node
        /// @return the data of the node
        /// @throws KeeperException If the server signals an error with a non-zero error code
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws IllegalArgumentException if an invalid path is specified
        /// </summary>
        public byte[] GetData(string path, IWatcher watcher, Stat stat)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            // the watch contains the un-chroot path
            WatchRegistration wcb = null;
            if (watcher != null)
            {
                wcb = new DataWatchRegistration(watchManager, watcher, clientPath);
            }

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.GetData;
            GetDataRequest request = new GetDataRequest();
            request.Path = serverPath;
            request.Watch = watcher != null;
            GetDataResponse response = new GetDataResponse();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, wcb);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            if (stat != null)
            {
                DataTree.CopyStat(response.Stat, stat);
            }
            return response.Data;
        }

        /// <summary>
        /// Return the data and the stat of the node of the given path.
        /// 
        /// If the watch is true and the call is successful (no exception is
        /// thrown), a watch will be left on the node with the given path. The watch
        /// will be triggered by a successful operation that sets data on the node, or
        /// deletes the node.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// @param path the given path
        /// @param watch whether need to watch this node
        /// @param stat the stat of the node
        /// @return the data of the node
        /// @throws KeeperException If the server signals an error with a non-zero error code
        /// @throws InterruptedException If the server transaction is interrupted.
        /// </summary>
        public byte[] GetData(string path, bool watch, Stat stat)
        {
            return GetData(path, watch ? watchManager.defaultWatcher : null, stat);
        }

        /// <summary>
        /// Set the data for the node of the given path if such a node exists and the
        /// given version matches the version of the node (if the given version is
        /// -1, it matches any node's versions). Return the stat of the node.
        /// 
        /// This operation, if successful, will trigger all the watches on the node
        /// of the given path left by getData calls.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// 
        /// A KeeperException with error code KeeperException.BadVersion will be
        /// thrown if the given version does not match the node's version.
        ///
        /// The maximum allowable size of the data array is 1 MB (1,048,576 bytes).
        /// Arrays larger than this will cause a KeeperExecption to be thrown.
        /// @param path
        ///                the path of the node
        /// @param data
        ///                the data to set
        /// @param version
        ///                the expected matching version
        /// @return the state of the node
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws KeeperException If the server signals an error with a non-zero error code.
        /// @throws IllegalArgumentException if an invalid path is specified
        /// </summary>
        public Stat SetData(string path, byte[] data, int version)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.SetData;
            SetDataRequest request = new SetDataRequest();
            request.Path = serverPath;
            request.Data = data;
            request.Version = version;
            SetDataResponse response = new SetDataResponse();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, null);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            return response.Stat;
        }

        /// <summary>
        /// Return the ACL and stat of the node of the given path.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// @param path
        ///                the given path for the node
        /// @param stat
        ///                the stat of the node will be copied to this parameter.
        /// @return the ACL array of the given node.
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws KeeperException If the server signals an error with a non-zero error code.
        /// @throws IllegalArgumentException if an invalid path is specified
        /// </summary>
        public List<ACL> GetACL(string path, Stat stat)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.GetACL;
            GetACLRequest request = new GetACLRequest();
            request.Path = serverPath;
            GetACLResponse response = new GetACLResponse();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, null);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            DataTree.CopyStat(response.Stat, stat);
            return response.Acl;
        }

        /// <summary>
        /// Set the ACL for the node of the given path if such a node exists and the
        /// given version matches the version of the node. Return the stat of the
        /// node.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// 
        /// A KeeperException with error code KeeperException.BadVersion will be
        /// thrown if the given version does not match the node's version.
        /// @param path
        /// @param acl
        /// @param version
        /// @return the stat of the node.
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws KeeperException If the server signals an error with a non-zero error code.
        /// @throws org.apache.zookeeper.KeeperException.InvalidACLException If the acl is invalide.
        /// @throws IllegalArgumentException if an invalid path is specified
        /// </summary>
        public Stat SetACL(string path, List<ACL> acl, int version)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.SetACL;
            SetACLRequest request = new SetACLRequest();
            request.Path = serverPath;
            if (acl != null && acl.Count == 0)
            {
                throw new KeeperException.InvalidACLException();
            }
            request.Acl = acl;
            request.Version = version;
            SetACLResponse response = new SetACLResponse();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, null);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            return response.Stat;
        }

        /// <summary>
        /// Return the list of the children of the node of the given path.
        /// 
        /// If the watch is non-null and the call is successful (no exception is thrown),
        /// a watch will be left on the node with the given path. The watch willbe
        /// triggered by a successful operation that deletes the node of the given
        /// path or creates/delete a child under the node.
        /// 
        /// The list of children returned is not sorted and no guarantee is provided
        /// as to its natural or lexical order.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// @param path
        /// @param watcher explicit watcher
        /// @return an unordered array of children of the node with the given path
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws KeeperException If the server signals an error with a non-zero error code.
        /// @throws IllegalArgumentException if an invalid path is specified
        /// </summary>
        public List<string> GetChildren(string path, IWatcher watcher)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            // the watch contains the un-chroot path
            WatchRegistration wcb = null;
            if (watcher != null)
            {
                wcb = new ChildWatchRegistration(watchManager, watcher, clientPath);
            }

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.GetChildren2;
            GetChildren2Request request = new GetChildren2Request();
            request.Path = serverPath;
            request.Watch = watcher != null;
            GetChildren2Response response = new GetChildren2Response();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, wcb);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            return response.Children;
        }

        public List<string> GetChildren(string path, bool watch)
        {
            return GetChildren(path, watch ? watchManager.defaultWatcher : null);
        }

        /// <summary>
        /// For the given znode path return the stat and children list.
        /// 
        /// If the watch is non-null and the call is successful (no exception is thrown),
        /// a watch will be left on the node with the given path. The watch willbe
        /// triggered by a successful operation that deletes the node of the given
        /// path or creates/delete a child under the node.
        /// 
        /// The list of children returned is not sorted and no guarantee is provided
        /// as to its natural or lexical order.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// @since 3.3.0
        /// 
        /// @param path
        /// @param watcher explicit watcher
        /// @param stat stat of the znode designated by path
        /// @return an unordered array of children of the node with the given path
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws KeeperException If the server signals an error with a non-zero error code.
        /// @throws IllegalArgumentException if an invalid path is specified
         /// </summary>

        public List<string> GetChildren(string path, IWatcher watcher, Stat stat)
        {
            string clientPath = path;
            PathUtils.ValidatePath(clientPath);

            // the watch contains the un-chroot path
            WatchRegistration wcb = null;
            if (watcher != null)
            {
                wcb = new ChildWatchRegistration(watchManager, watcher, clientPath);
            }

            string serverPath = PrependChroot(clientPath);

            RequestHeader h = new RequestHeader();
            h.Type = (int)OpCode.GetChildren2;
            GetChildren2Request request = new GetChildren2Request();
            request.Path = serverPath;
            request.Watch = watcher != null;
            GetChildren2Response response = new GetChildren2Response();
            ReplyHeader r = cnxn.SubmitRequest(h, request, response, wcb);
            if (r.Err != 0)
            {
                throw KeeperException.Create((KeeperException.Code)Enum.ToObject(typeof(KeeperException.Code), r.Err), clientPath);
            }
            if (stat != null)
            {
                DataTree.CopyStat(response.Stat, stat);
            }
            return response.Children;
        }

        /// <summary>
        /// For the given znode path return the stat and children list.
        /// 
        /// If the watch is true and the call is successful (no exception is thrown),
        /// a watch will be left on the node with the given path. The watch willbe
        /// triggered by a successful operation that deletes the node of the given
        /// path or creates/delete a child under the node.
        /// 
        /// The list of children returned is not sorted and no guarantee is provided
        /// as to its natural or lexical order.
        /// 
        /// A KeeperException with error code KeeperException.NoNode will be thrown
        /// if no node with the given path exists.
        /// @since 3.3.0
        /// 
        /// @param path
        /// @param watch
        /// @param stat stat of the znode designated by path
        /// @return an unordered array of children of the node with the given path
        /// @throws InterruptedException If the server transaction is interrupted.
        /// @throws KeeperException If the server signals an error with a non-zero
        ///  error code.
        /// </summary>
        public List<string> GetChildren(string path, bool watch, Stat stat)
        {
            return GetChildren(path, watch ? watchManager.defaultWatcher : null, stat);
        }

        /// <summary>
        /// string representation of this ZooKeeper client. Suitable for things
        /// like logging.
        /// 
        /// Do NOT count on the format of this string, it may change without
        /// warning.
        /// 
        /// @since 3.3.0
        /// </summary>
        public override string ToString()
        {
            return (string.Format("Id: {0} State:{1}{2}{3}", id, state, (state == States.CONNECTED ? string.Format(" Timeout:{0} ", SessionTimeout) : " "), cnxn));
        }
    }
}
